from typing import Dict
from unittest.mock import patch

import responses

from posit.connect import Client
from posit.connect.external.databricks import (
    CredentialsProvider,
    CredentialsStrategy,
    PositCredentialsProvider,
    PositCredentialsStrategy,
)


class mock_strategy(CredentialsStrategy):
    def auth_type(self) -> str:
        return "local"

    def __call__(self) -> CredentialsProvider:
        def inner() -> Dict[str, str]:
            return {"Authorization": "Bearer static-pat-token"}

        return inner


def register_mocks():
    responses.post(
        "https://connect.example/__api__/v1/oauth/integrations/credentials",
        match=[
            responses.matchers.urlencoded_params_matcher(
                {
                    "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                    "subject_token_type": "urn:posit:connect:user-session-token",
                    "subject_token": "cit",
                },
            ),
        ],
        json={
            "access_token": "dynamic-viewer-access-token",
            "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "token_type": "Bearer",
        },
    )


class TestPositCredentialsHelpers:
    @responses.activate
    def test_posit_credentials_provider(self):
        register_mocks()

        client = Client(api_key="12345", url="https://connect.example/")
        client._ctx.version = None
        cp = PositCredentialsProvider(client=client, user_session_token="cit")
        assert cp() == {"Authorization": "Bearer dynamic-viewer-access-token"}

    @responses.activate
    @patch.dict("os.environ", {"RSTUDIO_PRODUCT": "CONNECT"})
    def test_posit_credentials_strategy(self):
        register_mocks()

        client = Client(api_key="12345", url="https://connect.example/")
        client._ctx.version = None
        cs = PositCredentialsStrategy(
            local_strategy=mock_strategy(),
            user_session_token="cit",
            client=client,
        )
        cp = cs()
        assert cs.auth_type() == "posit-oauth-integration"
        assert cp() == {"Authorization": "Bearer dynamic-viewer-access-token"}

    def test_posit_credentials_strategy_fallback(self):
        # local_strategy is used when the content is running locally
        client = Client(api_key="12345", url="https://connect.example/")
        cs = PositCredentialsStrategy(
            local_strategy=mock_strategy(),
            user_session_token="cit",
            client=client,
        )
        cp = cs()
        assert cs.auth_type() == "local"
        assert cp() == {"Authorization": "Bearer static-pat-token"}



from posit import connect


class TestUser:
    @classmethod
    def setup_class(cls):
        cls.client = client = connect.Client()
        cls.aron = client.users.create(
            username="aron",
            email="aron@example.com",
            password="s3cur3p@ssword",
        )
        cls.bill = client.users.create(
            username="bill",
            email="bill@example.com",
            password="s3cur3p@ssword",
        )
        cls.cole = client.users.create(
            username="cole",
            email="cole@example.com",
            password="s3cur3p@ssword",
        )

    def test_lock(self):
        for user in (self.aron, self.bill, self.cole):
            user.lock()
            assert len(self.client.users.find(account_status="locked")) == 1
            user.unlock()
        assert len(self.client.users.find(account_status="locked")) == 0

    def test_count(self):
        # aron, bill, cole, and me
        assert self.client.users.count() == 4

    def test_find(self):
        assert self.client.users.find(prefix="aron") == [self.aron]
        assert self.client.users.find(prefix="bill") == [self.bill]
        assert self.client.users.find(prefix="cole") == [self.cole]

    def test_find_one(self):
        assert self.client.users.find_one(prefix="aron") == self.aron
        assert self.client.users.find_one(prefix="bill") == self.bill
        assert self.client.users.find_one(prefix="cole") == self.cole

    def test_get(self):
        assert self.client.users.get(self.aron["guid"]) == self.aron
        assert self.client.users.get(self.bill["guid"]) == self.bill
        assert self.client.users.get(self.cole["guid"]) == self.cole


class TestUserContent:
    """Checks behavior of the content attribute."""

    @classmethod
    def setup_class(cls):
        cls.client = client = connect.Client()
        cls.me = me = client.me
        cls.content = me.content.create(
            name="Sample",
            description="Simple sample content for testing",
            access_type="acl",
        )

    @classmethod
    def teardown_class(cls):
        assert cls.content.delete() is None
        assert cls.me.content.count() == 0

    def test_count(self):
        assert self.me.content.count() == 1

    def test_find(self):
        assert self.me.content.find()

    def test_find_one(self):
        assert self.me.content.find_one()

    def test_multiple_users(self):
        user = self.client.users.create(
            username="example",
            email="example@example.com",
            password="s3cur3p@ssword",
        )
        # assert filtering limits to the provided user.
        assert self.me.content.find_one() == self.content
        assert user.content.find_one() is None
        user.lock()



from __future__ import annotations

import posixpath
from urllib.parse import urlsplit, urlunsplit


class Url(str):
    """URL representation for Connect.

    An opinionated URL representation of a Connect URL. Maintains various
    conventions:
        - It begins with a scheme.
        - It is absolute.
        - It contains '__api__'.

    Supports Python builtin __add__ for append.

    Methods
    -------
    append(path: str)
        Append a path to the URL.

    Examples
    --------
    >>> url = Url("http://connect.example.com/")
    http://connect.example.com/__api__
    >>> url + "endpoint"
    http://connect.example.com/__api__/endpoint

    Append works with string-like objects (e.g., objects that support casting to string)
    >>> url = Url("http://connect.example.com/__api__/endpoint")
    http://connect.example.com/__api__/endpoint
    >>> url + 1
    http://connect.example.com/__api__/endpoint/1
    """

    def __new__(cls, value: str):
        url = _create(value)
        return super(Url, cls).__new__(cls, url)

    def __add__(self, path: str):
        return self.append(path)

    def append(self, path: str) -> Url:
        return Url(_append(self, path))


def _create(url: str) -> str:
    """Create a URL.

    Asserts that the URL is a proper Posit Connect endpoint. The path '__api__' is appended to the URL if it is missing.

    Parameters
    ----------
    url : str
        The original URL.

    Returns
    -------
    Url
        The validated and formatted URL.

    Raises
    ------
    ValueError
        The Url is missing a scheme.
    ValueError
        The Url is missing a network location (i.e., a domain name).

    Examples
    --------
    >>> _create("http://example.com")
    http://example.com/__api__

    >>> _create("http://example.com/__api__")
    http://example.com/__api__
    """
    split = urlsplit(url, allow_fragments=False)
    if not split.scheme:
        raise ValueError(f"URL must specify a scheme (e.g., http://example.com/__api__): {url}")
    if not split.netloc:
        raise ValueError(f"URL must be absolute (e.g., http://example.com/__api__): {url}")

    url = url.rstrip("/")
    if "/__api__" not in url:
        url = _append(url, "__api__")

    return url


def _append(url: str, path) -> str:
    """Append a path to a Url.

    Parameters
    ----------
    url : str
        A valid URL.
    path : str
        A valid path.

    Returns
    -------
    Url
        The original Url with the path appended to the end.

    Examples
    --------
    >>> url = _create("http://example.com/__api__")
    >>> _append(url, "path")
    http://example.com/__api__/path
    """
    path = str(path).strip("/")
    split = urlsplit(url, allow_fragments=False)
    new_path = posixpath.join(split.path, path)
    return urlunsplit((split.scheme, split.netloc, new_path, split.query, None))



import os


def is_local() -> bool:
    """Returns true if called from a piece of content running on a Connect server.

    The connect server will always set the environment variable `RSTUDIO_PRODUCT=CONNECT`.
    We can use this environment variable to determine if the content is running locally
    or on a Connect server.
    """
    return not os.getenv("RSTUDIO_PRODUCT") == "CONNECT"



from typing import List

from .resources import Resource, ResourceParameters, Resources
from .tasks import Task


class Variant(Resource):
    def render(self) -> Task:
        path = f"variants/{self['id']}/render"
        url = self.params.url + path
        response = self.params.session.post(url)
        return Task(self.params, **response.json())


class Variants(Resources):
    def __init__(self, params: ResourceParameters, content_guid: str) -> None:
        super().__init__(params)
        self.content_guid = content_guid

    def find(self) -> List[Variant]:
        path = f"applications/{self.content_guid}/variants"
        url = self.params.url + path
        response = self.params.session.get(url)
        results = response.json() or []
        return [Variant(self.params, **result) for result in results]



import functools
from typing import Optional, Protocol

import requests
from packaging.version import Version

from .urls import Url


def requires(version: str):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(instance: ContextManager, *args, **kwargs):
            ctx = instance._ctx
            if ctx.version and Version(ctx.version) < Version(version):
                raise RuntimeError(
                    f"This API is not available in Connect version {ctx.version}. Please upgrade to version {version} or later.",
                )
            return func(instance, *args, **kwargs)

        return wrapper

    return decorator


class Context(dict):
    def __init__(self, session: requests.Session, url: Url):
        self.session = session
        self.url = url

    @property
    def version(self) -> Optional[str]:
        try:
            value = self["version"]
        except KeyError:
            endpoint = self.url + "server_settings"
            response = self.session.get(endpoint)
            result = response.json()
            value = self["version"] = result.get("version")
        return value

    @version.setter
    def version(self, value):
        self["version"] = value


class ContextManager(Protocol):
    _ctx: Context



import pytest
from packaging import version

from posit import connect

from .. import CONNECT_VERSION


@pytest.mark.skipif(
    CONNECT_VERSION <= version.parse("2024.06.0"),
    reason="OAuth Integrations not supported.",
)
class TestIntegrations:
    @classmethod
    def setup_class(cls):
        cls.client = connect.Client()
        cls.integration = cls.client.oauth.integrations.create(
            name="example integration",
            description="integration description",
            template="custom",
            config={
                "auth_mode": "Confidential",
                "authorization_uri": "https://example.com/__tenand_id__/oauth2/v2.0/authorize",
                "client_id": "client_id",
                "client_secret": "client_secret",
                "scopes": "a b c",
                "token_endpoint_auth_method": "client_secret_post",
                "token_uri": "https://example.com/__tenant_id__/oauth2/v2.0/token",
            },
        )

        cls.another_integration = cls.client.oauth.integrations.create(
            name="another example integration",
            description="another integration description",
            template="custom",
            config={
                "auth_mode": "Confidential",
                "authorization_uri": "https://example.com/__tenand_id__/oauth2/v2.0/authorize",
                "client_id": "client_id",
                "client_secret": "client_secret",
                "scopes": "a b c",
                "token_endpoint_auth_method": "client_secret_post",
                "token_uri": "https://example.com/__tenant_id__/oauth2/v2.0/token",
            },
        )

    @classmethod
    def teardown_class(cls):
        cls.integration.delete()
        cls.another_integration.delete()
        assert len(cls.client.oauth.integrations.find()) == 0

    def test_get(self):
        result = self.client.oauth.integrations.get(self.integration["guid"])
        assert result == self.integration

    def test_find(self):
        results = self.client.oauth.integrations.find()
        assert len(results) == 2
        assert results[0] == self.integration
        assert results[1] == self.another_integration

    def test_create_update_delete(self):
        # create a new integration

        integration = self.client.oauth.integrations.create(
            name="new integration",
            description="new integration description",
            template="custom",
            config={
                "auth_mode": "Confidential",
                "authorization_uri": "https://example.com/__tenand_id__/oauth2/v2.0/authorize",
                "client_id": "client_id",
                "client_secret": "client_secret",
                "scopes": "a b c",
                "token_endpoint_auth_method": "client_secret_post",
                "token_uri": "https://example.com/__tenant_id__/oauth2/v2.0/token",
            },
        )

        created = self.client.oauth.integrations.get(integration["guid"])
        assert created == integration

        all_integrations = self.client.oauth.integrations.find()
        assert len(all_integrations) == 3

        # update the new integration

        created.update(name="updated integration name")
        updated = self.client.oauth.integrations.get(integration["guid"])
        assert updated["name"] == "updated integration name"

        # delete the new integration

        created.delete()
        all_integrations_after_delete = self.client.oauth.integrations.find()
        assert len(all_integrations_after_delete) == 2



from posit import connect


class TestEnvVars:
    @classmethod
    def setup_class(cls):
        cls.client = connect.Client()
        cls.content = cls.client.content.create(
            name="Sample",
            description="Simple sample content for testing",
            access_type="acl",
        )

    @classmethod
    def teardown_class(cls):
        cls.content.delete()
        assert cls.client.content.count() == 0

    def test_clear(self):
        self.content.environment_variables.create("KEY", "value")
        assert self.content.environment_variables.find() == ["KEY"]
        self.content.environment_variables.clear()
        assert self.content.environment_variables.find() == []

    def test_create(self):
        self.content.environment_variables.create("KEY", "value")
        assert self.content.environment_variables.find() == ["KEY"]

    def test_delete(self):
        self.content.environment_variables.create("KEY", "value")
        assert self.content.environment_variables.find() == ["KEY"]
        self.content.environment_variables.delete("KEY")
        assert self.content.environment_variables.find() == []

    def test_find(self):
        self.content.environment_variables.create("KEY", "value")
        assert self.content.environment_variables.find() == ["KEY"]

    def test_update(self):
        self.content.environment_variables.update(KEY="value")
        assert self.content.environment_variables.find() == ["KEY"]



from pathlib import Path

import pytest
from packaging import version

from posit import connect

from .. import CONNECT_VERSION


@pytest.mark.skipif(
    CONNECT_VERSION <= version.parse("2024.06.0"),
    reason="OAuth Integrations not supported.",
)
class TestAssociations:
    @classmethod
    def setup_class(cls):
        cls.client = connect.Client()
        cls.integration = cls.client.oauth.integrations.create(
            name="example integration",
            description="integration description",
            template="custom",
            config={
                "auth_mode": "Confidential",
                "authorization_uri": "https://example.com/__tenand_id__/oauth2/v2.0/authorize",
                "client_id": "client_id",
                "client_secret": "client_secret",
                "scopes": "a b c",
                "token_endpoint_auth_method": "client_secret_post",
                "token_uri": "https://example.com/__tenant_id__/oauth2/v2.0/token",
            },
        )

        cls.another_integration = cls.client.oauth.integrations.create(
            name="another example integration",
            description="another integration description",
            template="custom",
            config={
                "auth_mode": "Confidential",
                "authorization_uri": "https://example.com/__tenand_id__/oauth2/v2.0/authorize",
                "client_id": "client_id",
                "client_secret": "client_secret",
                "scopes": "a b c",
                "token_endpoint_auth_method": "client_secret_post",
                "token_uri": "https://example.com/__tenant_id__/oauth2/v2.0/token",
            },
        )

        # create content
        # requires full bundle deployment to produce an interactive content type
        cls.content = cls.client.content.create(name="example-flask-minimal")
        # create bundle
        path = Path("../../../../resources/connect/bundles/example-flask-minimal/bundle.tar.gz")
        path = (Path(__file__).parent / path).resolve()
        bundle = cls.content.bundles.create(str(path))
        # deploy bundle
        task = bundle.deploy()
        task.wait_for()

        cls.content.oauth.associations.update(cls.integration["guid"])

    @classmethod
    def teardown_class(cls):
        cls.integration.delete()
        cls.another_integration.delete()
        assert len(cls.client.oauth.integrations.find()) == 0

        cls.content.delete()
        assert cls.client.content.count() == 0

    def test_find_by_integration(self):
        associations = self.integration.associations.find()
        assert len(associations) == 1
        assert associations[0]["oauth_integration_guid"] == self.integration["guid"]

        no_associations = self.another_integration.associations.find()
        assert len(no_associations) == 0

    def test_find_update_by_content(self):
        associations = self.content.oauth.associations.find()
        assert len(associations) == 1
        assert associations[0]["app_guid"] == self.content["guid"]
        assert associations[0]["oauth_integration_guid"] == self.integration["guid"]

        # update content association to another_integration
        self.content.oauth.associations.update(self.another_integration["guid"])
        updated_associations = self.content.oauth.associations.find()
        assert len(updated_associations) == 1
        assert updated_associations[0]["app_guid"] == self.content["guid"]
        assert (
            updated_associations[0]["oauth_integration_guid"] == self.another_integration["guid"]
        )

        # unset content association
        self.content.oauth.associations.delete()
        no_associations = self.content.oauth.associations.find()
        assert len(no_associations) == 0



import io
from unittest.mock import Mock

import pytest
import responses
from requests import HTTPError, JSONDecodeError, Response

from posit.connect import Client
from posit.connect.errors import ClientError
from posit.connect.hooks import handle_errors


def test_success():
    response = Mock()
    response.status_code = 200
    assert handle_errors(response) == response


def test_client_error():
    response = Mock()
    response.status_code = 400
    response.json = Mock(return_value={"code": 0, "error": "foobar"})
    with pytest.raises(ClientError):
        handle_errors(response)


def test_client_error_without_payload():
    class StatusException(Exception):
        pass

    response = Mock()
    response.status_code = 404
    response.json = Mock(side_effect=JSONDecodeError("Test code", "Test msg", 0))
    response.raise_for_status = Mock(side_effect=StatusException())
    with pytest.raises(StatusException):
        handle_errors(response)


def test_200():
    response = Response()
    response.status_code = 200
    assert handle_errors(response) == response


def test_response_client_error_with_plaintext_payload():
    response = Response()
    response.status_code = 404
    response.raw = io.BytesIO(b"Plain text 404 Not Found")
    with pytest.raises(HTTPError):
        handle_errors(response)


def test_response_client_error_with_json_payload():
    response = Response()
    response.status_code = 400
    response.raw = io.BytesIO(b'{"code":0,"error":"foobar"}')
    with pytest.raises(
        ClientError,
        match='{"error_code": 0, "error_message": "foobar", "http_status": 400, "http_message": "Bad Request", "payload": null}',
    ):
        handle_errors(response)


def test_response_client_error_without_payload():
    response = Response()
    response.status_code = 404
    response.raw = io.BytesIO(b"Plain text 404 Not Found")
    with pytest.raises(HTTPError):
        handle_errors(response)


@responses.activate
def test_deprecation_warning():
    responses.get(
        "https://connect.example/__api__/v0",
        headers={"X-Deprecated-Endpoint": "v1/"},
    )
    c = Client("https://connect.example", "12345")

    with pytest.warns(DeprecationWarning):
        c.get("v0")



from __future__ import annotations

import posixpath
from typing import Generator, Literal, Optional, TypedDict

from typing_extensions import NotRequired, Required, Unpack

from posit.connect.context import requires
from posit.connect.errors import ClientError
from posit.connect.paginator import Paginator

from .resources import Active, ActiveFinderMethods, ActiveSequence


class ContentPackage(Active):
    class _Package(TypedDict):
        language: Required[str]
        name: Required[str]
        version: Required[str]
        hash: Required[Optional[str]]

    def __init__(self, ctx, /, **attributes: Unpack[_Package]):
        # todo - passing "" is a hack since path isn't needed. Instead, this class should inherit from Resource, but ActiveSequence is designed to operate on Active. That should change.
        super().__init__(ctx, "", **attributes)


class ContentPackages(ActiveFinderMethods["ContentPackage"], ActiveSequence["ContentPackage"]):
    """A collection of packages."""

    def __init__(self, ctx, path):
        super().__init__(ctx, path, "name")

    def _create_instance(self, path, /, **attributes):  # noqa: ARG002
        return ContentPackage(self._ctx, **attributes)

    def fetch(self, **conditions):
        try:
            return super().fetch(**conditions)
        except ClientError as e:
            if e.http_status == 204:
                return []
            raise e

    def find(self, uid):
        raise NotImplementedError("The 'find' method is not support by the Packages API.")

    class _FindBy(TypedDict, total=False):
        language: NotRequired[Literal["python", "r"]]
        """Programming language ecosystem, options are 'python' and 'r'"""

        name: NotRequired[str]
        """The package name"""

        version: NotRequired[str]
        """The package version"""

        hash: NotRequired[Optional[str]]
        """Package description hash for R packages."""

    def find_by(self, **conditions: Unpack[_FindBy]):  # type: ignore
        """
        Find the first record matching the specified conditions.

        There is no implied ordering, so if order matters, you should specify it yourself.

        Parameters
        ----------
        **conditions : Unpack[_FindBy]
            Conditions for filtering packages. The following keys are accepted:

        language : {"python", "r"}, not required
            Programming language ecosystem, options are 'python' and 'r'

        name : str, not required
            The package name

        version : str, not required
            The package version

        hash : str or None, optional, not required
            Package description hash for R packages.

        Returns
        -------
        Optional[T]
            The first record matching the specified conditions, or `None` if no such record exists.
        """
        return super().find_by(**conditions)


class ContentPackagesMixin(Active):
    """Mixin class to add a packages attribute."""

    @property
    @requires(version="2024.10.0-dev")
    def packages(self):
        path = posixpath.join(self._path, "packages")
        return ContentPackages(self._ctx, path)


class Package(Active):
    class _Package(TypedDict):
        language: Required[Literal["python", "r"]]
        """Programming language ecosystem, options are 'python' and 'r'"""

        language_version: Required[str]
        """Programming language version"""

        name: Required[str]
        """The package name"""

        version: Required[str]
        """The package version"""

        hash: Required[Optional[str]]
        """Package description hash for R packages."""

        bundle_id: Required[str]
        """The unique identifier of the bundle this package is associated with"""

        app_id: Required[str]
        """The numerical identifier of the application this package is associated with"""

        app_guid: Required[str]
        """The unique identifier of the application this package is associated with"""

    def __init__(self, ctx, /, **attributes: Unpack[_Package]):
        # todo - passing "" is a hack since path isn't needed. Instead, this class should inherit from Resource, but ActiveSequence is designed to operate on Active. That should change.
        super().__init__(ctx, "", **attributes)


class Packages(ActiveFinderMethods["Package"], ActiveSequence["Package"]):
    def __init__(self, ctx, path):
        super().__init__(ctx, path, "name")

    def _create_instance(self, path, /, **attributes):  # noqa: ARG002
        return Package(self._ctx, **attributes)

    class _Fetch(TypedDict, total=False):
        language: Required[Literal["python", "r"]]
        """Programming language ecosystem, options are 'python' and 'r'"""

        name: Required[str]
        """The package name"""

        version: Required[str]
        """The package version"""

    def fetch(self, **conditions: Unpack[_Fetch]) -> Generator["Package"]:  # type: ignore
        # todo - add pagination support to ActiveSequence
        url = self._ctx.url + self._path
        paginator = Paginator(self._ctx.session, url, dict(**conditions))
        for page in paginator.fetch_pages():
            results = page.results
            yield from (self._create_instance("", **result) for result in results)

    def find(self, uid):
        raise NotImplementedError("The 'find' method is not support by the Packages API.")

    class _FindBy(TypedDict, total=False):
        language: NotRequired[Literal["python", "r"]]
        """Programming language ecosystem, options are 'python' and 'r'"""

        language_version: NotRequired[str]
        """Programming language version"""

        name: NotRequired[str]
        """The package name"""

        version: NotRequired[str]
        """The package version"""

        hash: NotRequired[Optional[str]]
        """Package description hash for R packages."""

        bundle_id: NotRequired[str]
        """The unique identifier of the bundle this package is associated with"""

        app_id: NotRequired[str]
        """The numerical identifier of the application this package is associated with"""

        app_guid: NotRequired[str]
        """The unique identifier of the application this package is associated with"""

    def find_by(self, **conditions: Unpack[_FindBy]) -> "Package | None":  # type: ignore
        """
        Find the first record matching the specified conditions.

        There is no implied ordering, so if order matters, you should specify it yourself.

        Parameters
        ----------
        **conditions : Unpack[_FindBy]
            Conditions for filtering packages. The following keys are accepted:

        language : {"python", "r"}, not required
            Programming language ecosystem, options are 'python' and 'r'

        name : str, not required
            The package name

        version : str, not required
            The package version

        hash : str or None, optional, not required
            Package description hash for R packages.

        Returns
        -------
        Optional[Package]
            The first record matching the specified conditions, or `None` if no such record exists.
        """
        return super().find_by(**conditions)
